
class ClusteringExperiment:

    def __init__(self,exp_id="Unnamed",data=None,regressor=None,aux_data=None):

        self.id=exp_id
        self.data=data
        self.regressor=regressor
        self.aux_data=aux_data

        self._confirm_time_coords_match()
        self.regression_correlations=None
        self.regressed_pcs=None
        self.windowed_pcs=None
        self.windowed_regressed_pcs=None
        self.windowed_field_data=None
        self.windowed_regressor=None
        self.clusters={}
        self.cluster_cubes={}
        self.cluster_correlations={}

    #Not currently very rigorous.
    #Makes sure the time axes align.
    def _confirm_time_coords_match(self):

        t_axes=[]
        if self.pcs is not None:
            t_axes.append(self.pcs.coord("time"))
        if self.regressor is not None:
            t_axes.append(self.regressor.coord("time"))
        if self.field_data is not None:
            t_axes.append(self.field_data.coord("time"))

        for t_ax in t_axes:
            assert np.all(t_ax.points==t_axes[0].points)

    #regresses self.regressor against self.pcs
    def regress_pcs(self):

        if self.regressor is None:
            raise(ValueError("No regressor attribute defined."))


        corr,regressed_pc=regress_pc(self.pcs,self.regressor.data)

        self.regression_correlations=corr
        self.regressed_pcs=regressed_pc

    #Called by combine_with, and used to stick different
    #PC sequences together.
    def _combine_attribute(self,attr,cluster_array,time_coord=None):

        array=[getattr(self,attr).data]

        for C in cluster_array:
            array.append(getattr(C,attr).data)

        #The atleast_3d here helps make sure 1D attributes
        #(like the regressor) get treated the same way as 2D attributes
        array=np.vstack(np.atleast_3d(array))
        T=array.shape[0]

        new_attr=make_cube_with_different_1st_axis(getattr(self,attr),T,t_ax=time_coord)
        #We want to get rid of length 1 dimensions, hence the squeeze here
        new_attr.data=np.squeeze(array)
        return new_attr

    #Combines the current ClusteringExperiment PCs with some others,
    #appending state sequences together.
    def combine_with(self,cluster_array,time_coord=None,new_id=None):

        New_ClusterExperiment=ClusteringExperiment(exp_id=new_id)

        for attribute in ["pcs","regressed_pcs","regressor","field_data"]:
            if getattr(self,attribute) is not None:
                combined_attribute=self._combine_attribute(attribute,cluster_array,time_coord)
                setattr(New_ClusterExperiment,attribute,combined_attribute)

        return New_ClusterExperiment

    def _window_attribute(self,attribute,width,overlap):

        data=getattr(self,attribute)

        windowed_array=[]

        T=data.shape[0]
        window_num=np.floor((T-width)/overlap).astype(int)
        windows=[slice(i*overlap,(i*overlap)+width) for i in range(window_num)]

        for window in windows:
            windowed_array.append(data[window])

        return windowed_array

    def window_data(self,width,overlap):

        for attribute in ["pcs","regressed_pcs","regressor","field_data"]:
            if getattr(self,attribute) is not None:
                windowed_attribute=self._window_attribute(attribute,width,overlap)
                setattr(self,"windowed_"+attribute,windowed_attribute)


    def cluster_pcs(self,Ks,pc_list=None):

        self.Ks=Ks

        if pc_list is None:
            pc_list=["pcs","regressed_pcs","windowed_pcs","windowed_regressed_pcs"]

        for pcs in pc_list:
            if getattr(self,pcs) is not None:
                clusters=self._cluster_attribute(pcs,Ks)
                self.clusters[pcs]=clusters

    def _cluster_attribute(self,pcs,Ks):

        data=getattr(self,pcs)
        #data will either be an iris cube or a list of cubes.
        #We try and iterate and if that fails then its a cube

        try:
            clusters=[{K:Kmeans_cluster(cube.data,K) for K in Ks} for cube in data]
        except:
            clusters={K:Kmeans_cluster(data.data,K) for K in Ks}

        return clusters

    def get_cluster_cubes(self,pc_list=None):

        if pc_list is None:
            pc_list=["pcs","regressed_pcs","windowed_pcs","windowed_regressed_pcs"]

        for pc in pc_list:

            if getattr(self,pc) is not None:

                cl_data=self.clusters[pc]
                #Assume its iterable (i.e. windowed):
                try:
                    ccs=[{K:get_cluster_cube(F,cl[K].states) for K in self.Ks} for F,cl in zip(self.windowed_field_data,cl_data)]
                #If not, then its full data:
                except:

                    ccs={K:get_cluster_cube(self.field_data,cl_data[K].states) for K in self.Ks}

                self.cluster_cubes[pc]=ccs

    def _reorder_clusters(self,pc,K,mapping,window=None):

        order=np.array([m[1] for m in mapping])

        if window is None:
            #Reorder cluster cubes
            self.cluster_cubes[pc][K]=self.cluster_cubes[pc][K][order]
            #Reorder cluster data
            self.clusters[pc][K].reorder(mapping)
        else:
            #Reorder cluster cubes
            self.cluster_cubes[pc][window][K]=self.cluster_cubes[pc][window][K][order]
            #Reorder cluster data
            self.clusters[pc][window][K].reorder(mapping)


    def correlate_clusters(self,reference_clusters,reorder_clusters=False,pc_list=None,reference_id="None"):

        if pc_list is None:
            pc_list=["pcs","regressed_pcs","windowed_pcs","windowed_regressed_pcs"]

        correlation_dict={}

        for pc in pc_list:

            cubes1=self.cluster_cubes[pc]


            #Non windowed clusters:
            if type(cubes1) is dict:

                mean_corr_dict={}
                reg_corr_dict={}
                for K in self.Ks:

                    (mean_corr,reg_corrs),mapping=correlate_clusters(reference_clusters[K],cubes1[K],and_mapping=True,mean_only=False)

                    mean_corr_dict[K]=mean_corr
                    reg_corr_dict[K]=reg_corrs

                    if reorder_clusters:
                        self._reorder_clusters(pc,K,mapping,window=None)

                correlation_dict[pc]=[mean_corr_dict,reg_corr_dict]

            #Windowed clusters
            elif type(cubes1) is list:

                mean_corr_arr=[]
                reg_corr_arr=[]

                for w,cc1 in enumerate(cubes1):

                    mean_corr_dict={}
                    reg_corr_dict={}
                    for K in self.Ks:

                        (mean_corr,reg_corrs),mapping=correlate_clusters(reference_clusters[K],cc1[K],and_mapping=True,mean_only=False)

                        mean_corr_dict[K]=mean_corr
                        reg_corr_dict[K]=reg_corrs

                        if reorder_clusters:
                            self._reorder_clusters(pc,K,mapping,window=w)

                    mean_corr_arr.append(mean_corr_dict)
                    reg_corr_arr.append(reg_corr_dict)

                correlation_dict[pc]=[mean_corr_arr,reg_corr_arr]

            else:
                raise(ValueError("cubes should be stored either as a dict or list"))

        self.cluster_correlations[reference_id]=correlation_dict

    #Just a little syntactic sugar to make data retrieval more convenient

    def get_cl_data(self,pc_data,cl_attr,K,window=None):
        if window is None:
            return getattr(self.clusters[pc_data][K],cl_attr)
        else:
            return getattr(self.clusters[pc_data][window][K],cl_attr)

    def get_cc_data(self,pc_data,K,window=None):
        if window is None:
            return self.cluster_cubes[pc_data][K]
        else:
            return self.cluster_cubes[pc_data][window][K]

    def get_correlation_data(self,exp_id,pc_data,K,window):

        if window is None:
            return self.cluster_correlations[exp_id][pc_data][K]
        else:
            return self.cluster_correlations[exp_id][pc_data][window][K]




import os
import xarray as xr
import numpy as np
import cftime
from deseasonaliser import Agg_Deseasonaliser, Sinefit_Deseasonaliser

class Preprocessor(object):
    """
    Handles the initial processing of datasets. All functionality can be used for
    a single input dataset or a dict of datasets, which will be either xarray
    DataArrays or iris Cubes, and will be converted to DataArrays on loading.
    ----------
    Methods
    -------
    align(data_keys=None,**kwargs): A wrapper around xarray.align that applies
    to the datasets specified by data_keys, or to all datasets if data_keys is None.
    kwargs pass directly to xarray.align.

    subset(coords,coord_bounds,data_keys=None): coords takes a list of N
    strings specifying coords to be subsetted. coord_bounds is a Nx2 list of
    numerical data which specifies the lower and upper bounds (inclusively)
     for each coord.

    squeeze(data_keys=None): wrapper round xarray.squeeze
    detrend(self,data_keys=None,dim="time",deg=1): wrapper round xarray.polyfit
    and polyval which removes a polynomial trend from the given data_keys.

    fit_seasonal_cycle(data_keys=None,deseasonaliser=Agg_Deseasonaliser,
    **ds_kwargs): Calculates coeffs needed to evaluate seasonal cycles
     given an input time coordinate.

    evaluate_seasonal_cycle(data_key=None): Returns a seasonal cycle
     DataArray for the given DataArray

    deseasonalise(data_keys=None,cycle=None,compute_cycle=True): By default
     removes the seasonal cycle of every dataset from itself, using
    default kwargs of evaluate_seasonal_cycle if the cycle has not been computed.
    If compute_cycle=False will instead return error if cycle does not exist.
    If cycle is a single value then that cycle will be removed from all data_keys.
    """
    def __init__(self,data,id="unnamed"):

        self.id=id

        #If not a dict make it a dict
        if type(data)!=dict:
            data={"input":data}

        self.data=data

        self._check_data_types()
        self._create_data_attributes()
        self.deseasonalisers={}

    def _check_data_types(self):
        datarray_type=type(xr.DataArray())
        try:
            self.data={m:x if type(x)==datarray_type else xr.DataArray.from_iris(x)\
             for m,x in self.data.items()}

        except:
            raise(IOError("All input data must be xarray DataArrays or iris Cubes"))

    def _create_data_attributes(self):

        for m in self.data:
            setattr(self,m,self.data[m])
        return

    def __repr__(self):
        return f"Preprocessor object: {self.id}"

    def __str__(self):
        out=[self.__repr__()]
        for m,x in self.data.items():
            out.append(f"\n{m}: {x.shape}")
        return "".join(out)

    def _update_data_attr(self, attr,new_val):
        self.data[attr]=new_val
        setattr(self,attr,new_val)
        return

    #Align data
    def align(self,data_keys=None,**kwargs):

        #if data is None we align all the data
        #otherwise just the named variables

        if data_keys is None:
            data_keys=list(self.data.keys())

        in_data=[self.data[d] for d in np.atleast_1d(data_keys)]

        aligned_data=xr.align(*in_data,**kwargs)
        for m,x in zip(data_keys,aligned_data):
            self._update_data_attr(m,x)

    def subset(self,coords,coord_bounds,data_keys=None):
            if data_keys is None:
                data_keys=list(self.data.keys())

            coords=np.atleast_1d(coords)
            coord_bounds=np.atleast_2d(coord_bounds)
            subsetted_data={}

            for key in data_keys:
                filter=None

                for coord,(b0,b1) in zip(coords,coord_bounds):
                    #Is true when the coord is between the inclusive limits
                    condition=(self.data[key].coords[coord]>=b0)&(self.data[key].coords[coord]<=b1)
                    #multiply each filter together
                    if filter is None:
                        filter=condition
                    else:
                        filter=filter*condition

                subsetted_data[key]=self.data[key].where(filter,drop=True)

            #We dont change any data until the whole function
            #runs without errors.
            for m,x in subsetted_data.items():
                self._update_data_attr(m,x)
            return
    def squeeze(self,data_keys=None):
        if data_keys is None:
            data_keys=list(self.data.keys())

        for m in data_keys:
            self._update_data_attr(m,self.data[m].squeeze())
        return


    def _detrend_arr(self,data,dim,deg):

        p = data.polyfit(dim=dim, deg=deg)
        fit = xr.polyval(data[dim], p.polyfit_coefficients)
        detrended=data-fit
        return detrended

    def detrend(self,data_keys=None,dim="time",deg=1):
        if data_keys is None:
            data_keys=list(self.data.keys())
        for m in data_keys:
            self._update_data_attr(m,self._detrend_arr(self.data[m],dim,deg))
        return

    def fit_seasonal_cycle(self,data_keys=None,deseasonaliser=Agg_Deseasonaliser,**ds_kwargs):
        if data_keys is None:
            data_keys=list(self.data.keys())
        for m in data_keys:
            dsnlsr=deseasonaliser()
            dsnlsr.fit_cycle(self.data[m])
            self.deseasonalisers[m]=dsnlsr

        return

    def evaluate_seasonal_cycle(self,data_key=None):

        try:
            dsnlsr=self.deseasonalisers[data_key]
        except KeyError as err:
            try:
                dsnlsr=self.deseasonalisers["input"]
            except:
                raise(err)

        return dsnlsr.evaluate_cycle()

    #By default removes the seasonal cycle of every dataset from itself, using
    #default kwargs of get_seasonal_cycle if the cycle has not been computed.
    #If compute_cycle=False will instead return error if cycle does not exist.
    #If cycle is a single value then that cycle will be removed from all data_keys.
    def deseasonalise(self,data_keys=None,cycle=None,compute_cycle=True):
        if data_keys is None:
            data_keys=list(self.data.keys())
        if cycle is None:
            cycle=data_keys
        elif type(cycle)==str:
            cycle=np.repeat(cycle,len(data_keys))

        for c,m in zip(cycle,data_keys):
            try:
                dsnlsr=self.deseasonalisers[c]
            except KeyError as err:
                if compute_cycle:
                    self.fit_seasonal_cycle(data_keys=[c])
                    dsnlsr=self.deseasonalisers[c]
                else:
                    raise(err)

            deseasonalised_data=self.data[m]
            cycle=dsnlsr.evaluate_cycle(data=self.data[m])
            deseasonalised_data.data=deseasonalised_data.data-cycle.data
            self._update_data_attr(m,deseasonalised_data)
        return

    #TO DO

    #Interpolating data to match
    #Take EOFs
    #Match time coords



import numpy as np
import xarray as xr
import cftime
import pandas as pd
from scipy.optimize import leastsq
import os
import matplotlib.pyplot as plt
import time
import cftime
from util import xarr_times_to_ints

class Agg_Deseasonaliser(object):
    """
    Computes a seasonal cycle from an xarray dataset based on an aggregation
    criteria i.e. month, day of year, week of year etc.
    ----------
    Methods
    -------
    fit_cycle(DataArray,dim=str,agg=str): Aggregates the input DataArray arr.

    evaluate_cycle(data=DataArray): When data is None, creates a seasonal cycle
    DataArray the same shape as the input data used in fit_cycle. data can be
    a DataArray with a differing dim coordinate, but matching other coordinates.
    This can be used to fit a seasonal cycle on one dataset, but evaluate it on
    another, as long as the grid, levels, etc are the same.
    Returns The seasonal cycle as a DataArray
    """
    def __init__(self):
        self.data=None
        self.cycle_coeffs=None


    def fit_cycle(self,arr,dim="time",agg="dayofyear"):
        var=dim+"."+agg
        self.data=arr
        self.dim=dim
        self.agg=agg
        self.cycle_coeffs=arr.groupby(var).mean(dim=dim)
        return

    def evaluate_cycle(self,data=None):
        if data is None:
            data=self.data
        return self.cycle_coeffs[getattr(data[self.dim].dt,self.agg).data-1]

class Sinefit_Deseasonaliser(Agg_Deseasonaliser):

    """
    Same functionalityy as Agg_Deseasonaliser but fits a number of sin modes
    to the data using least squares fitting. This produces a smoother cycle than
    aggregating but can be slow for large datasets.
    ----------
    Methods
    -------
    fit_cycle(DataArray,dim=str,N=int,period=float): N sets the number of sin modes
    to fit as fractions of the period. i.e. the default of N=4 fits waves with
    periods of period, period/2, period/3 and period/4. period is a float setting
    the length of a year in days.

    evaluate_cycle(data=DataArray): As for Agg_Deseasonaliser
    """
    def _func_residual(self,p,x,t,N,period):
        return x - self._evaluate_fit(t,p,N,period)

    def _evaluate_fit(self,t,p,N,period):
        ans=p[1]
        for i in range(0,N):
            ans+=p[2*i+2] * np.sin(2 * np.pi * (i+1)/period * t + p[2*i+3])
        return ans

    def _lstsq_sine_fit(self,arr,t,N,period):

        #Guess initial parameters
        std=arr.std()
        p=np.zeros(2*(N+1))
        for i in range(0,N):
            p[2+2*i]=std/(i+1.0)
        plsq=leastsq(self._func_residual,p,args=(arr,t,N,period))
        return plsq[0]

    def fit_cycle(self,arr,dim="time",N=4,period=365.25):

            dims=arr.dims
            self.dim=dim
            self.data=arr
            self.N=N
            self.period=period
            t=xarr_times_to_ints(arr[dim])
            self.coeffs= xr.apply_ufunc(
                self._lstsq_sine_fit,
                arr,
                input_core_dims=[[dim]],
                output_core_dims=[["coeffs"]],
                vectorize=True,
                kwargs={"t":t,"N":N,"period":period})
            return

    def evaluate_cycle(self,data=None):
        if data is None:
            data=self.data
        dims=data.dims
        t=xarr_times_to_ints(data[self.dim])

        cycle=np.array([[self._evaluate_fit(t,c2.data,self.N,self.period)\
                for c2 in c1] for c1 in self.coeffs])

        return data.transpose(...,"time").copy(data=cycle).transpose(*dims)
